<?php

namespace Drupal\k8s\Controller;

use Drupal\cloud\Entity\CloudProjectInterface;
use Drupal\Component\Serialization\Yaml;
use Drupal\Core\Controller\ControllerBase;
use Drupal\Core\Database\Connection;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Link;
use Drupal\Core\Messenger\Messenger;
use Drupal\Core\Render\RendererInterface;
use Drupal\Core\Url;
use Drupal\k8s\Entity\K8sNode;
use Drupal\k8s\Entity\K8sPod;
use Drupal\k8s\Service\K8sService;
use Drupal\k8s\Service\K8sServiceInterface;
use Drupal\k8s\Traits\K8sBlockTrait;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\RequestStack;

/**
 * Controller responsible for "update" urls.
 *
 * This class is mainly responsible for
 * updating the K8s entities from urls.
 */
class ApiController extends ControllerBase implements ApiControllerInterface {

  use K8sBlockTrait;

  /**
   * Drupal\Core\Entity\EntityTypeManagerInterface definition.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The K8s Service.
   *
   * @var \Drupal\k8s\Service\K8sServiceInterface
   */
  private $k8sService;

  /**
   * The Messenger service.
   *
   * @var \Drupal\Core\Messenger\Messenger
   */
  protected $messenger;

  /**
   * Request stack.
   *
   * @var \Symfony\Component\HttpFoundation\RequestStack
   */
  private $requestStack;

  /**
   * Renderer service.
   *
   * @var \Drupal\Core\Render\RendererInterface
   */
  private $renderer;

  /**
   * The database service.
   *
   * @var \Drupal\Core\Database\Connection
   */
  private $database;

  /**
   * ApiController constructor.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\k8s\Service\K8sServiceInterface $k8s_service
   *   Object for interfacing with K8s API.
   * @param \Drupal\Core\Messenger\Messenger $messenger
   *   Messanger Object.
   * @param \Symfony\Component\HttpFoundation\RequestStack $request_stack
   *   Request stack object.
   * @param \Drupal\Core\Render\RendererInterface $renderer
   *   The renderer service.
   * @param \Drupal\Core\Database\Connection $database
   *   The database service.
   */
  public function __construct(
    EntityTypeManagerInterface $entity_type_manager,
    K8sServiceInterface $k8s_service,
    Messenger $messenger,
    RequestStack $request_stack,
    RendererInterface $renderer,
    Connection $database
  ) {
    $this->entityTypeManager = $entity_type_manager;
    $this->k8sService = $k8s_service;
    $this->messenger = $messenger;
    $this->requestStack = $request_stack;
    $this->renderer = $renderer;
    $this->database = $database;
  }

  /**
   * Dependency Injection.
   *
   * @param \Symfony\Component\DependencyInjection\ContainerInterface $container
   *   Instance of ContainerInterface.
   *
   * @return ApiController
   *   return created object.
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('entity_type.manager'),
      $container->get('k8s'),
      $container->get('messenger'),
      $container->get('request_stack'),
      $container->get('renderer'),
      $container->get('database')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function updateAllResources($cloud_context) {
    k8s_update_resources($cloud_context);

    return $this->redirect('view.k8s_node.list', [
      'cloud_context' => $cloud_context,
    ]);
  }

  /**
   * {@inheritdoc}
   */
  public function updateNodeList($cloud_context = '') {
    return $this->updateEntityList('node', $cloud_context);
  }

  /**
   * {@inheritdoc}
   */
  public function updateNamespaceList($cloud_context = '') {
    return $this->updateEntityList('namespace', $cloud_context);
  }

  /**
   * {@inheritdoc}
   */
  public function updatePodList($cloud_context = '') {
    return $this->updateEntityList('pod', $cloud_context);
  }

  /**
   * {@inheritdoc}
   */
  public function updateNetworkPolicyList($cloud_context = '') {
    return $this->updateEntityList('network_policy', $cloud_context);
  }

  /**
   * {@inheritdoc}
   */
  public function updateDeploymentList($cloud_context = '') {
    return $this->updateEntityList('deployment', $cloud_context);
  }

  /**
   * {@inheritdoc}
   */
  public function updateReplicaSetList($cloud_context = '') {
    return $this->updateEntityList('replica_set', $cloud_context);
  }

  /**
   * {@inheritdoc}
   */
  public function updateServiceList($cloud_context = '') {
    return $this->updateEntityList('service', $cloud_context);
  }

  /**
   * {@inheritdoc}
   */
  public function updateCronJobList($cloud_context = '') {
    return $this->updateEntityList('cron_job', $cloud_context);
  }

  /**
   * {@inheritdoc}
   */
  public function updateJobList($cloud_context = '') {
    return $this->updateEntityList('job', $cloud_context);
  }

  /**
   * {@inheritdoc}
   */
  public function updateResourceQuotaList($cloud_context = '') {
    return $this->updateEntityList('resource_quota', $cloud_context);
  }

  /**
   * {@inheritdoc}
   */
  public function updateLimitRangeList($cloud_context = '') {
    return $this->updateEntityList('limit_range', $cloud_context);
  }

  /**
   * {@inheritdoc}
   */
  public function updateSecretList($cloud_context = '') {
    return $this->updateEntityList('secret', $cloud_context);
  }

  /**
   * {@inheritdoc}
   */
  public function updateConfigMapList($cloud_context = '') {
    return $this->updateEntityList('config_map', $cloud_context);
  }

  /**
   * {@inheritdoc}
   */
  public function updateRoleList($cloud_context = '') {
    return $this->updateEntityList('role', $cloud_context);
  }

  /**
   * {@inheritdoc}
   */
  public function updateClusterRoleList($cloud_context = '') {
    return $this->updateEntityList('cluster_role', $cloud_context);
  }

  /**
   * {@inheritdoc}
   */
  public function updatePersistentVolumeList($cloud_context = '') {
    return $this->updateEntityList('persistent_volume', $cloud_context);
  }

  /**
   * {@inheritdoc}
   */
  public function updateStorageClassList($cloud_context = '') {
    return $this->updateEntityList('storage_class', $cloud_context);
  }

  /**
   * {@inheritdoc}
   */
  public function updateStatefulSetsList($cloud_context = '') {
    return $this->updateEntityList('stateful_set', $cloud_context);
  }

  /**
   * {@inheritdoc}
   */
  public function updateIngressList($cloud_context = '') {
    return $this->updateEntityList('ingress', $cloud_context);
  }

  /**
   * {@inheritdoc}
   */
  public function updateDaemonSetList($cloud_context = '') {
    return $this->updateEntityList('daemon_set', $cloud_context);
  }

  /**
   * {@inheritdoc}
   */
  public function updateEndpointList($cloud_context = '') {
    return $this->updateEntityList('endpoint', $cloud_context);
  }

  /**
   * {@inheritdoc}
   */
  public function updateEventList($cloud_context = '') {
    return $this->updateEntityList('event', $cloud_context);
  }

  /**
   * {@inheritdoc}
   */
  public function updatePersistentVolumeClaimList($cloud_context = '') {
    return $this->updateEntityList('persistent_volume_claim', $cloud_context);
  }

  /**
   * {@inheritdoc}
   */
  public function updateClusterRoleBindingList($cloud_context = '') {
    return $this->updateEntityList('cluster_role_binding', $cloud_context);
  }

  /**
   * {@inheritdoc}
   */
  public function updateServiceAccountsList($cloud_context = '') {
    return $this->updateEntityList('service_account', $cloud_context);
  }

  /**
   * {@inheritdoc}
   */
  public function updateApiServiceList($cloud_context = '') {
    return $this->updateEntityList('api_service', $cloud_context);
  }

  /**
   * {@inheritdoc}
   */
  public function updateRoleBindingsList($cloud_context = '') {
    return $this->updateEntityList('role_binding', $cloud_context);
  }

  /**
   * {@inheritdoc}
   */
  public function updatePriorityClassesList($cloud_context = '') {
    return $this->updateEntityList('priority_class', $cloud_context);
  }

  /**
   * {@inheritdoc}
   */
  public function getNodeMetrics($cloud_context, K8sNode $k8s_node) {
    $node_name = $k8s_node->getName();

    $result = $this->database
      ->select('watchdog', 'w')
      ->fields('w', ['message', 'timestamp'])
      ->condition('type', 'k8s_metrics')
      ->condition('timestamp', time() - 7 * 24 * 60 * 60, '>=')
      ->condition('message', "%$node_name%", 'LIKE')
      ->orderBy('timestamp')
      ->execute();

    $data = [];
    foreach ($result ?: [] as $record) {
      $message = Yaml::decode($record->message);
      if (empty($message[$cloud_context])) {
        continue;
      }

      $cpu = NULL;
      if (empty($message[$cloud_context]['nodes'][$node_name])) {
        continue;
      }

      $cpu = $message[$cloud_context]['nodes'][$node_name]['cpu'];
      $memory = $message[$cloud_context]['nodes'][$node_name]['memory'] ?? 0;
      $data[] = [
        'timestamp' => $record->timestamp,
        'cpu' => $cpu,
        'memory' => $memory,
      ];
    }

    return new JsonResponse($data);
  }

  /**
   * {@inheritdoc}
   */
  public function getPodMetrics($cloud_context, K8sPod $k8s_pod) {
    $pod_name = $k8s_pod->getName();
    $pod_namespace = $k8s_pod->getNamespace();

    $result = $this->database
      ->select('watchdog', 'w')
      ->fields('w', ['message', 'timestamp'])
      ->condition('type', 'k8s_metrics')
      ->condition('timestamp', time() - 7 * 24 * 60 * 60, '>=')
      ->condition('message', "%$pod_namespace:$pod_name%", 'LIKE')
      ->orderBy('timestamp')
      ->execute();

    $data = [];
    foreach ($result ?: [] as $record) {
      $message = Yaml::decode($record->message);
      if (empty($message[$cloud_context])) {
        continue;
      }

      $cpu = NULL;
      $key = "$pod_namespace:$pod_name";
      if (empty($message[$cloud_context]['pods'][$key])) {
        continue;
      }

      $cpu = $message[$cloud_context]['pods'][$key]['cpu'];
      $memory = $message[$cloud_context]['pods'][$key]['memory'] ?? 0;
      $data[] = [
        'timestamp' => $record->timestamp,
        'cpu' => $cpu,
        'memory' => $memory,
      ];
    }

    return new JsonResponse($data);
  }

  /**
   * Get all metrics information on all nodes in a cluster of $cloud_context.
   *
   * @param string $cloud_context
   *   The cloud context.
   *
   * @return \Symfony\Component\HttpFoundation\JsonResponse
   *   All metrics information on all nodes across a cluster specified by
   *   $cloud_context.
   */
  public function getNodeAllocatedResourcesList($cloud_context) {
    return $this->getNodeAllocatedResources($cloud_context);
  }

  /**
   * Get the metrics information on all node(s)
   *
   * @return \Symfony\Component\HttpFoundation\JsonResponse
   *   The metrics information on node(s).
   */
  public function getAllNodeAllocatedResources(): JsonResponse {
    return new JsonResponse($this->loadAllNodeAllocatedResources());
  }

  /**
   * Get the metrics information on the node(s)
   *
   * @param string $cloud_context
   *   The cloud context.
   * @param \Drupal\cloud\Entity\CloudProjectInterface $cloud_project
   *   The cloud project.
   * @param \Drupal\k8s\Entity\K8sNode $k8s_node
   *   The K8s node entity to expect to get the metrics information.  If this
   *   parameter is omitted, it assumes that all nodes are specified.
   *
   * @return \Symfony\Component\HttpFoundation\JsonResponse
   *   The metrics information on node(s).
   */
  public function getNodeAllocatedResources($cloud_context, CloudProjectInterface $cloud_project = NULL, K8sNode $k8s_node = NULL): JsonResponse {
    return new JsonResponse($this->loadNodeAllocatedResources($cloud_context, $cloud_project, $k8s_node));
  }

  /**
   * Helper method to update entities.
   *
   * @param string $entity_type_name
   *   The entity type name.
   * @param string $cloud_context
   *   The cloud context.
   *
   * @return \Symfony\Component\HttpFoundation\RedirectResponse
   *   An associative array with a redirect route and any parameters to build
   *   the route.
   */
  private function updateEntityList($entity_type_name, $cloud_context) {
    if (empty($cloud_context)) {
      return $this->updateAllEntityList($entity_type_name);
    }

    $labels = $this->getDisplayLabels('k8s_' . $entity_type_name);
    $update_method_name = 'update' . self::getCamelCaseWithoutWhitespace($labels['plural']);
    $this->k8sService->setCloudContext($cloud_context);
    $updated = $this->k8sService->$update_method_name();

    if ($updated !== FALSE) {
      $this->messageUser($this->t('Updated @name.', ['@name' => $labels['plural']]));
      K8sService::clearCacheValue();
    }
    else {
      $this->messageUser($this->t('Unable to update @name.', ['@name' => $labels['plural']]), 'error');
    }

    // Update the cache.
    K8sService::clearCacheValue();

    return $this->redirect("view.k8s_$entity_type_name.list", [
      'cloud_context' => $cloud_context,
    ]);
  }

  /**
   * Helper method to update all entities.
   *
   * @param string $entity_type_name
   *   The entity type name.
   *
   * @return \Symfony\Component\HttpFoundation\RedirectResponse
   *   An associative array with a redirect route and any parameters to build
   *   the route.
   */
  private function updateAllEntityList($entity_type_name) {
    $cloud_configs = $this->entityTypeManager
      ->getStorage('cloud_config')
      ->loadByProperties([
        'type' => 'k8s',
      ]);
    $labels = $this->getDisplayLabels('k8s_' . $entity_type_name);
    $update_method_name = 'update' . self::getCamelCaseWithoutWhitespace($labels['plural']);

    foreach ($cloud_configs ?: [] as $cloud_config) {
      $cloud_context = $cloud_config->getCloudContext();
      $this->k8sService->setCloudContext($cloud_context);
      $updated = $this->k8sService->$update_method_name();

      $resource_link = Link::fromTextAndUrl(
        $labels['plural'],
        Url::fromRoute(
          "entity.k8s_${entity_type_name}.collection",
          ['cloud_context' => $cloud_context]
        )
      )->toString();

      $cloud_config_link = $cloud_config->getCloudConfigLink();

      if ($updated !== FALSE) {
        $this->messageUser($this->t('Updated %resource_link of %cloud_config_link cloud service provider.', [
          '%resource_link' => $resource_link ?? $labels['plural'],
          '%cloud_config_link' => $cloud_config_link ?? $cloud_config->getName(),
        ]));
      }
      else {
        $this->messageUser($this->t('Unable to update %resource_link of %cloud_config_link cloud service provider.', [
          '%resource_link' => $resource_link ?? $labels['plural'],
          '%cloud_config_link' => $cloud_config_link ?? $cloud_config->getName(),
        ]), 'error'
        );
      }
    }

    // Update the cache.
    K8sService::clearCacheValue();

    return $this->redirect("view.k8s_$entity_type_name.all");
  }

  /**
   * Helper method to add messages for the end user.
   *
   * @param string $message
   *   The message.
   * @param string $type
   *   The message type: error or message.
   */
  private function messageUser($message, $type = 'message') {
    switch ($type) {
      case 'error':
        $this->messenger->addError($message);
        break;

      case 'message':
        $this->messenger->addStatus($message);
      default:
        break;
    }
  }

}
