<?php

namespace Drupal\k8s\Controller;

use Drupal\Component\Serialization\Yaml;
use Drupal\cloud\Entity\CloudProjectInterface;
use Drupal\k8s\Entity\K8sNodeInterface;
use Symfony\Component\HttpFoundation\JsonResponse;

/**
 * Controller responsible for K8s Node Allocated Resources Chart.
 */
class K8sNodeAllocatedResourcesChartController extends K8sCostsControllerBase {

  /**
   * Get allocated resources for all node.
   *
   * @return \Symfony\Component\HttpFoundation\JsonResponse
   *   A JSON response of a K8s Node Allocated Resources.
   */
  public function getAllK8sNodeAllocatedResources() {
    $node_ids = $this->entityTypeManager()
      ->getStorage('k8s_node')
      ->getQuery()
      ->execute();
    $nodes = $this->entityTypeManager()
      ->getStorage('k8s_node')->loadMultiple($node_ids);

    return $this->calculateNodeAllocatedResources($nodes);
  }

  /**
   * Get K8s Node Allocated Resources.
   *
   * @param string $cloud_context
   *   Cloud context string.
   * @param \Drupal\cloud\Entity\CloudProjectInterface $cloud_project
   *   The cloud project.
   * @param \Drupal\k8s\Entity\K8sNodeInterface $k8s_node
   *   K8s Node entity.
   *
   * @return \Symfony\Component\HttpFoundation\JsonResponse
   *   A JSON response of a K8s Node Allocated Resources.
   */
  public function getK8sNodeAllocatedResources($cloud_context, CloudProjectInterface $cloud_project = NULL, K8sNodeInterface $k8s_node = NULL) {
    $nodes = [];
    $cloud_contexts = [];

    if (!empty($cloud_project)) {
      $k8s_clusters = $cloud_project->get('field_k8s_clusters')->getValue();
      foreach ($k8s_clusters ?: [] as $k8s_cluster) {
        $cloud_contexts[] = $k8s_cluster['value'];
      }
    }
    elseif (!empty($cloud_context)) {
      $cloud_contexts[] = $cloud_context;
    }

    // If cloud_context is passed, filter entities with it.
    if (!empty($cloud_contexts)) {
      foreach ($cloud_contexts ?: [] as $cloud_context) {
        $nodes = array_merge($nodes, isset($k8s_node)
          ? $this->entityTypeManager()->getStorage('k8s_node')
            ->loadByProperties([
              'id' => $k8s_node->id(),
            ])
          : $this->entityTypeManager()->getStorage('k8s_node')
            ->loadByProperties([
              'cloud_context' => $cloud_context,
            ]));
      }
    }
    return !empty($cloud_project)
      ? $this->calculateNodeAllocatedResources($nodes)
      : $this->calculateNodeAllocatedResources($nodes, $cloud_context);
  }

  /**
   * Calculate node allocated resources.
   *
   * @param array $nodes
   *   Array of K8s Node entities.
   * @param string $cloud_context
   *   Optional cloud context.
   *
   * @return \Symfony\Component\HttpFoundation\JsonResponse
   *   Json response.
   */
  private function calculateNodeAllocatedResources(array $nodes, $cloud_context = NULL) {

    $period = $this->request->get('period');

    if (!isset($period) || empty($period)) {
      $period = 14;
    }

    $node_metrics = $this->getAllNodesMetrics($cloud_context, $period);
    $response = [];

    foreach ($node_metrics ?: [] as $time => $node_datas) {
      $cpu_usage = (float) 0;
      $memory_usage = (float) 0;
      $pod_usage = 0;
      $cpu_capacity = (float) 0;
      $memory_capacity = (float) 0;
      $pod_capacity = 0;
      $data = [];
      foreach ($nodes ?: [] as $node) {
        $key = $node->getName();
        if (isset($node_datas[$key])) {
          $node_data = $node_datas[$key];
          $cpu_usage += $node_data['cpu'];
          $memory_usage += $node_data['memory'];
          $pod_usage += $node_data['pod_allocation'];
          $cpu_capacity += $node_data['cpu_capacity'];
          $memory_capacity += $node_data['memory_capacity'];
          $pod_capacity += $node_data['pod_capacity'];
        }
      }
      $data['cpu'] = [
        'timestamp' => $time,
        'cpu_usage' => $cpu_usage,
        'cpu_capacity' => $cpu_capacity,
      ];
      $data['memory'] = [
        'timestamp' => $time,
        'memory_usage' => $memory_usage,
        'memory_capacity' => $memory_capacity,
      ];
      $data['pod'] = [
        'timestamp' => $time,
        'pod_usage' => $pod_usage,
        'pod_capacity' => $pod_capacity,
      ];
      $response[] = $data;
    }

    return new JsonResponse($response);
  }

  /**
   * Get all nodes metrics.
   *
   * @param string $cloud_context
   *   Cloud context string.  If no cloud context passed, retrieve
   *   all metric messages.
   * @param int $period
   *   The period to get data.
   *
   * @return array
   *   The node metrics array.
   */
  private function getAllNodesMetrics($cloud_context = NULL, $period = 14) {
    $query = $this->database
      ->select('watchdog', 'w')
      ->fields('w', ['message', 'timestamp'])
      ->condition('type', 'k8s_metrics')
      ->condition('timestamp', time() - (int) $period * 24 * 60 * 60, '>=');

    // Add cloud_context if set.
    if (!empty($cloud_context)) {
      $query->condition('message', "%$cloud_context:%", 'LIKE');
    }

    $result = $query->condition('message', '%nodes:%', 'LIKE')
      ->orderBy('timestamp')
      ->execute();
    $node_metrics = [];
    foreach ($result ?: [] as $record) {
      $message = Yaml::decode($record->message);
      if (!empty($cloud_context)) {
        if (empty($message[$cloud_context])) {
          continue;
        }
        if (empty($message[$cloud_context]['nodes'])) {
          continue;
        }
        $nodes = $message[$cloud_context];
      }
      else {
        $nodes = array_shift($message);
      }
      $date = date('Y-m-d H:i', $record->timestamp);
      $timestamp = strtotime($date);
      if (!isset($node_metrics[$timestamp])) {
        $node_metrics[$timestamp] = [];
      }
      foreach ($nodes['nodes'] ?: [] as $key => $node_data) {
        if (isset($node_data['cpu_capacity']) && isset($node_data['memory_capacity']) && isset($node_data['pod_capacity'])) {
          $node_metrics[$timestamp][$key] = [
            'cpu' => $node_data['cpu'],
            'cpu_capacity' => $node_data['cpu_capacity'],
            'memory' => $node_data['memory'] ?? 0,
            'memory_capacity' => $node_data['memory_capacity'],
            'pod_allocation' => $node_data['pod_allocation'],
            'pod_capacity' => $node_data['pod_capacity'],
          ];
        }
      }
    }

    return $node_metrics;
  }

}
