<?php

namespace Drupal\k8s\Controller;

use Drupal\Component\Plugin\Exception\PluginNotFoundException;
use Drupal\cloud\Entity\CloudProjectInterface;
use Symfony\Component\HttpFoundation\JsonResponse;

/**
 * Controller responsible for K8s Project Costs Chart.
 */
class K8sProjectCostsChartController extends K8sCostsControllerBase {

  /**
   * Get cost for all projects.
   *
   * @return \Symfony\Component\HttpFoundation\JsonResponse
   *   A JSON response of a K8s Project Costs.
   */
  public function getAllK8sProjectCosts() {
    $storage = $this->entityTypeManager()->getStorage('cloud_project');
    $cloud_projects = $storage->loadByProperties(['type' => 'k8s']);
    return $this->calculateProjectCosts($cloud_projects, TRUE);
  }

  /**
   * Get cost for all projects with clusters.
   *
   * @return \Symfony\Component\HttpFoundation\JsonResponse
   *   A JSON response of a K8s Project Costs.
   */
  public function getAllK8sProjectClusterCosts() {
    $storage = $this->entityTypeManager()->getStorage('cloud_project');
    $cloud_projects = $storage->loadByProperties(['type' => 'k8s']);
    return $this->calculateProjectCosts($cloud_projects, TRUE, TRUE);
  }

  /**
   * Get K8s Project Costs.
   *
   * @param string $cloud_context
   *   Cloud context string.
   *
   * @return \Symfony\Component\HttpFoundation\JsonResponse
   *   A JSON response of a K8s Project Costs.
   */
  public function getK8sProjectCosts($cloud_context) {
    $storage = $this->entityTypeManager()->getStorage('cloud_project');
    $cloud_projects = $storage->loadByProperties([
      'cloud_context' => $cloud_context,
      'type' => 'k8s',
    ]);
    return $this->calculateProjectCosts($cloud_projects, TRUE, FALSE);
  }

  /**
   * Get K8s Project Costs with clusters.
   *
   * @param string $cloud_context
   *   Cloud context string.
   * @param \Drupal\cloud\Entity\CloudProjectInterface $cloud_project
   *   The cloud project.
   *
   * @return \Symfony\Component\HttpFoundation\JsonResponse
   *   A JSON response of a K8s Project Costs.
   */
  public function getK8sProjectClusterCosts($cloud_context, CloudProjectInterface $cloud_project = NULL) {
    $cloud_projects = [];

    if (!empty($cloud_project)) {
      $cloud_projects[] = $cloud_project;
    }
    elseif (!empty($cloud_context)) {
      $storage = $this->entityTypeManager()->getStorage('cloud_project');
      $cloud_projects = $storage->loadByProperties([
        'cloud_context' => $cloud_context,
        'type' => 'k8s',
      ]);
    }
    return !empty($cloud_project)
      ? $this->calculateProjectCosts($cloud_projects, FALSE, TRUE)
      : $this->calculateProjectCosts($cloud_projects, TRUE, TRUE);
  }

  /**
   * Calculate node allocated resources.
   *
   * @param array $cloud_projects
   *   Array of K8s Cloud Project entities.
   * @param bool $need_project
   *   Whether needs project.
   * @param bool $need_cluster
   *   Whether needs clusters.
   *
   * @return \Symfony\Component\HttpFoundation\JsonResponse
   *   Json response.
   */
  private function calculateProjectCosts(array $cloud_projects, $need_project = TRUE, $need_cluster = FALSE) {
    $response = [];
    $project_costs = $this->getProjectCosts($cloud_projects, $need_project, $need_cluster);
    foreach ($project_costs ?: [] as $project => $costs) {
      $data = [
        'project' => "$project ",
      ];
      foreach ($costs as $time => $cost) {
        $data['costs'][] = [
          'timestamp' => $time,
          'cost' => $cost,
        ];
      }
      $response[] = $data;
    }
    return new JsonResponse($response);
  }

  /**
   * Get project costs from resource storage.
   *
   * @param array $cloud_projects
   *   Array of K8s Cloud Project entities.
   * @param bool $need_project
   *   Whether needs project.
   * @param bool $need_cluster
   *   Whether needs clusters.
   *
   * @return array
   *   The pod metrics array.
   */
  private function getProjectCosts(array $cloud_projects, $need_project = TRUE, $need_cluster = FALSE): array {
    if (!$this->moduleHandler()->moduleExists('cloud_budget')) {
      return [];
    }
    $plugin = NULL;
    try {
      $plugin_manager = \Drupal::service('plugin.manager.cloud_cost_storage');
      $plugin = $plugin_manager->createInstance('K8s_cloud_cost_storage');
    }
    catch (PluginNotFoundException $e) {
      return [];
    }

    $cost_type = $this->request->get('cost_type');
    if (!isset($cost_type) || empty($cost_type)) {
      $cost_type = 'ri_one_year';
    }

    $period = $this->request->get('period');
    if (!isset($period) || empty($period)) {
      $period = 14;
    }

    $from_time = time() - (int) $period * 24 * 60 * 60;
    $to_time = time();
    $project_costs = [];
    foreach ($cloud_projects ?: [] as $cloud_project) {
      $k8s_clusters = $cloud_project->get('field_k8s_clusters')->getValue();
      foreach ($k8s_clusters ?: [] as $k8s_cluster) {
        $k8s_namespaces = $this->entityTypeManager()
          ->getStorage('k8s_namespace')->loadByProperties([
            'cloud_context' => $k8s_cluster['value'],
          ]
        );

        $project = '';
        if ($need_project) {
          $project = $cloud_project->getName();
        }
        if ($need_cluster) {
          $project .= $need_project ? ':' . $k8s_cluster['value'] : $k8s_cluster['value'];
        }

        foreach ($k8s_namespaces ?: [] as $k8s_namespace) {
          $result = $plugin->extract('cloud_resource_storage', $k8s_namespace->getCloudContext(), $k8s_namespace->getName(), $from_time, $to_time, $cost_type);

          if (count($result) > 1) {
            if (!isset($project_costs[$project])) {
              $project_costs[$project] = [];
            }
            foreach ($result as $key => $record) {
              if ($key !== 'total_cost') {
                $date = date('Y-m-d H:i', $record['created']);
                $timestamp = strtotime($date);
                if (!isset($project_costs[$project][$timestamp])) {
                  $project_costs[$project][$timestamp] = $record['cost'];
                }
                else {
                  $project_costs[$project][$timestamp] += $record['cost'];
                }
              }
            }
          }
        }
      }
    }

    return $project_costs;
  }

  /**
   * {@inheritdoc}
   */
  public function getEc2ChartPeriod($cloud_context = NULL, CloudProjectInterface $cloud_project = NULL, $json = NULL) {
    $period_option = [
      1 => $this->t('One day'),
      7 => $this->t('One week'),
      14 => $this->t('Two weeks'),
      31 => $this->t('One month'),
      183 => $this->t('Half a year'),
      365 => $this->t('One year'),
    ];
    if (!$this->moduleHandler()->moduleExists('cloud_budget')) {
      if (isset($json)) {
        return new JsonResponse($period_option);
      }
      else {
        return $period_option;
      }
    }
    $cloud_contexts = [];
    $k8s_namespaces = [];
    $timestamp = 0;

    if (!empty($cloud_project)) {
      $k8s_clusters = $cloud_project->get('field_k8s_clusters')->getValue();
      foreach ($k8s_clusters ?: [] as $k8s_cluster) {
        $cloud_contexts[] = $k8s_cluster['value'];
      }
    }
    elseif (!empty($cloud_context)) {
      $cloud_contexts[] = $cloud_context;
    }
    foreach ($cloud_contexts ?: [] as $cluster) {
      $namespaces = $this->entityTypeManager()
        ->getStorage('k8s_namespace')->loadByProperties([
          'cloud_context' => $cluster,
        ]
      );
      if (!empty($namespaces)) {
        foreach ($namespaces ?: [] as $namespace) {
          $k8s_namespaces[] = $namespace->getName();
        }
      }
    }

    try {
      $query = $this->database->select('cloud_resource_storage', 'crs');
      $query->addExpression('min(crs.created)');
      if (!empty($cloud_contexts)) {
        $query->condition('cloud_context', $cloud_contexts, 'IN');
      }
      if (!empty($k8s_namespaces)) {
        $query->condition('payer', $k8s_namespaces, 'IN');
      }
      $timestamp = $query->execute()->fetchField();
    }
    catch (\Exception $e) {
      $this->handleException($e);
    }

    $keys = array_keys($period_option);
    $keys = array_reverse($keys);
    foreach ($keys ?: [] as $idx => $day) {
      if ($day > 7 && (time() - $day * 24 * 60 * 60 <= $timestamp) && (time() - $keys[$idx + 1] * 24 * 60 * 60 <= $timestamp)) {
        unset($period_option[$day]);
      }
    }
    if (isset($json)) {
      return new JsonResponse($period_option);
    }
    else {
      return $period_option;
    }
  }

}
